# -*- coding: utf-8 -*-

"""
This defines the base "simmate" command that all other commands stem from.
"""

import logging
from pathlib import Path

import typer

from simmate.command_line.database import database_app
from simmate.command_line.utilities import utilities_app
from simmate.command_line.workflow_engine import workflow_engine_app
from simmate.command_line.workflows import workflows_app

simmate_app = typer.Typer(rich_markup_mode="markdown")


@simmate_app.callback(no_args_is_help=True)
def base_command():
    """
    This is the base command that all other Simmate commands stem from
    :fire::fire::rocket:

    ---------

    If you are a beginner to the command line, be sure to start with
    [our tutorials](https://jacksund.github.io/simmate/getting_started/overview/).
    Below you will see a list of sub-commands to try. For example, you can run `simmate
    database --help` to learn more about it.

    \n
    TIP: Many Simmate commands are long and verbose. You can use `--install-completion`
    to add ipython-like autocomplete to your shell.

    """
    # When we call the command "simmate" this is where we start, and it then
    # looks for all other functions that have the decorator "@simmate.command()"
    # to decide what to do from there.
    pass


@simmate_app.command()
def run_server():
    """
    Runs a local test server for the Simmate website interface

    While this command is running, you can then view the working website
    at http://127.0.0.1:8000/

    This command is exactly the same as running:

    `django runserver --settings=simmate.configuration.django.settings`
    """

    # BUG: logging prints ugly duplicates as it sets up simmate and also a
    # static server for it.
    logging.warning(
        "Seeing duplicate messages is normal and expected. "
        "This is because the test server sets up + tears down Simmate on a cycle."
    )

    from django.core.management import call_command

    from simmate.database import connect

    call_command("runserver")


@simmate_app.command()
def start_project(project_name: Path):
    """
    Creates a new folder and fills it with an example project to
    get you started with custom Simmate workflows/datatables
    """

    import logging

    from django.core.management import call_command

    from simmate import configuration  # needed just for the filepath

    # This directory is where our template folder is located. We find this
    # by looking at the import path to see where python installed it.
    config_directory = Path(configuration.__file__).absolute().parent

    # We add on "project_template" to this file path as that's our full template
    template_directory = config_directory / "example_project"

    # we now make the project folder using our template directory.
    # Note, we are using Django's "startproject" command even though we are just
    # copying files over. This might be overkill but it gets the job done.
    call_command(
        "startproject",
        project_name,
        template=str(template_directory),
        extensions=["toml", "py"],
    )

    # grab the full path to the new project for the user to see
    new_project_directory = Path.cwd() / project_name

    # also navigate to the user's ~/simmate/applications.yaml file and we
    # want to add a new line at the end of it (or create the file if it isn't
    # there yet)
    apps_yaml = Path.home() / "simmate" / "applications.yaml"
    new_line = "\nexample_app.apps.ExampleAppConfig"  # \n ensures a new line

    # If the file exists, we append this line to the end of the file. Otherwise,
    # we create a new file and add the line!
    with apps_yaml.open("a+") as file:
        file.write(new_line)
    # Let the user know what we did and how to continue.
    logging.info(
        "\n\tSuccessfully made a new project! \n\t"
        f"You'll find it at {new_project_directory}\n\n"
        "\tBe sure to go through the README file in your new project.\n"
    )


# # All commands are organized into other files, so we need to manually register
# # them to our base "simmate" command here.
simmate_app.add_typer(database_app, name="database")
simmate_app.add_typer(workflows_app, name="workflows")
simmate_app.add_typer(workflow_engine_app, name="workflow-engine")
simmate_app.add_typer(utilities_app, name="utilities")



# -*- coding: utf-8 -*-

import itertools
import logging
from pathlib import Path

from rich.progress import track

from simmate.toolkit import Composition
from simmate.toolkit.structure_prediction import (
    get_known_structures,
    get_structures_from_prototypes,
)
from simmate.toolkit.structure_prediction.evolution.workflows.fixed_composition import (
    StructurePrediction__Python__FixedComposition,
)
from simmate.utilities import get_directory
from simmate.workflow_engine import Workflow
from simmate.workflows.utilities import get_workflow

# TODO
# StructurePrediction__Python__VariableTernaryComposition
#   --> calls VariableBinaryComposition strategically
#   --> might call FixedCompositionVariableNsites strategically too


class StructurePrediction__Python__BinaryComposition(Workflow):

    use_database = False

    fixed_comp_workflow = StructurePrediction__Python__FixedComposition

    @classmethod
    def run_config(
        cls,
        chemical_system: str,
        max_atoms: int = 10,
        subworkflow_name: str = "relaxation.vasp.staged",
        subworkflow_kwargs: dict = {},
        # max_stoich_factor: int = 4,
        directory: Path = None,
        singleshot_sources: list[str] = [],
        **kwargs,  # passed to fixed_comp_workflow
    ):

        # ---------------------------------------------------------------------
        # Setting up
        # ---------------------------------------------------------------------

        subworkflow = get_workflow(subworkflow_name)

        # Grab the two elements that we are working with. Also if this raises
        # an error, we want to provide useful feedback to the user
        try:
            element_1, element_2 = chemical_system.split("-")
        except:
            raise Exception(
                "Failed to split {} into elements. Your format for the input "
                "chemical system should be two elements and follow the format "
                "Element1-Element2. Ex: Na-Cl, Ca-N, Y-C, etc."
            )

        # Find all unique compositions
        compositions = []
        for nsites in range(1, max_atoms + 1):
            for combo in itertools.combinations_with_replacement(
                [element_1, element_2], nsites
            ):

                composition = Composition(
                    {
                        element_1: combo.count(element_1),
                        element_2: combo.count(element_2),
                    }
                )

                compositions.append(composition)
        logging.info(f"{len(compositions)} unique compositions will be explored")

        # now condense this list down to just the reduced formulas
        compositions_reduced = []
        for composition in compositions:
            reduced = composition.reduced_composition
            if reduced not in compositions_reduced:
                compositions_reduced.append(reduced)
        logging.info(
            f"This corresponds to {len(compositions_reduced)} reduced compositions"
        )

        # it is also useful to store the "max compositions". This the composition
        # for when the max number of possible sites are used. For example,
        # Ca2N with max atoms of 20 would give a max composition of Ca12N6 (18 atoms).
        # We make this list using only the reduced compositions as input.
        compositions_maxed = []
        for composition in compositions_reduced:
            max_factor = int(max_atoms // composition.num_atoms)
            # BUG: we need to call Composition() to convert the pymatgen
            # object back into a simmate composition.
            max_composition = Composition(max_factor * composition)

            if max_composition not in compositions_maxed:
                compositions_maxed.append(max_composition)

        # ---------------------------------------------------------------------
        # Submitting known structures
        # ---------------------------------------------------------------------

        logging.info("Generating input structures from third-party databases")
        structures_known = []
        for composition in track(compositions_maxed):
            new_structures = get_known_structures(
                composition,
                allow_multiples=True,
                nsites__lte=max_atoms,
            )
            structures_known += new_structures
        logging.info(
            f"Generated {len(structures_known)} structures from other databases"
        )

        # sort the structures from fewest nsites to most so that we can submit
        # them in this order.
        structures_known.sort(key=lambda s: s.num_sites)

        # write cif files
        directory_known = get_directory(directory / "known_structures")
        for i, s in enumerate(structures_known):
            s.to("cif", directory_known / f"{i}.cif")

        # and submit them and disable the logs while we submit
        logging.info("Submitting known structures")
        logger = logging.getLogger()
        logger.disabled = True
        for structure in track(structures_known):
            subworkflow.run_cloud(
                structure=structure,
                **subworkflow_kwargs,
            )
        logger.disabled = False

        # ---------------------------------------------------------------------
        # Submitting structures from prototypes
        # ---------------------------------------------------------------------

        # Start by generating the singleshot sources for each factor size.
        logging.info("Generating input structures from prototypes")
        structures_prototype = []
        # for singleshot_source in singleshot_sources: ## TODO
        for composition in track(compositions_maxed):

            # generate all prototypes
            new_structures = get_structures_from_prototypes(
                composition,
                max_sites=max_atoms,
            )
            structures_prototype += new_structures
        logging.info(
            f"Generated {len(structures_prototype)} structures from prototypes"
        )

        # sort the structures from fewest nsites to most so that we can submit
        # them in this order.
        structures_prototype.sort(key=lambda s: s.num_sites)

        directory_sub = get_directory(directory / "from_prototypes")
        for i, s in enumerate(structures_prototype):
            s.to("cif", directory_sub / f"{i}.cif")

        # and submit them and disable the logs while we submit
        logging.info("Submitting prototype structures")
        logger = logging.getLogger()
        logger.disabled = True
        for structure in track(structures_prototype):
            subworkflow.run_cloud(
                structure=structure,
                **subworkflow_kwargs,
            )
        logger.disabled = False

        # ---------------------------------------------------------------------
        # Starting search
        # ---------------------------------------------------------------------

        # Now we are going to cycle through factor sizes, starting at the lowest
        # and working our way up. We do this because:
        #   1. smaller unitcells run much faster (for DFT-calculation)
        #   2. smaller compositions converge much faster (for number of
        #       calcs required)
        #   3. results from smaller unitcells can help to speed up the larger
        #       compositional searches.

        for natoms in range(1, max_atoms + 1):

            logging.info(f"Beginning compositional searches with natoms = {natoms}")
            current_compositions = [c for c in compositions if c.num_atoms == natoms]

            for composition in current_compositions:

                # OPTIMIZE: Should I just skip single-element compositions?
                # BUG: I do skip these for now because compositions like "N1"
                # are failing in vasp for me.
                if len(composition.elements) == 1:
                    logging.warn(f"Skipping single-element composition {composition}")
                    continue

                # OPTIMIZE: Setting stop conditions for each search will be
                # essential in ensure we don't waste too much calculation time
                # on one composition -- while simultaniously making sure we
                # searched a compositon enough.
                n = composition.num_atoms
                min_structures_exact = int(10 * n)
                limit_best_survival = int(20 * n)
                max_structures = int(30 * n)

                # NOTES: This code below is for printing out the cutoff limits
                # I keep this here for early development as we figure out ideal
                # stopping conditions.
                # for n in range(1, 10):
                #     min_structures_exact = int(10 * n)
                #     limit_best_survival = int(10 * n)
                #     max_structures = int(30 * n)
                #     print(
                #         f"{n}\t{min_structures_exact}\t"
                #         f"{limit_best_survival}\t{max_structures}"
                #     )

                cls.fixed_comp_workflow.run(
                    composition=composition,
                    subworkflow_name=subworkflow_name,
                    subworkflow_kwargs=subworkflow_kwargs,
                    min_structures_exact=min_structures_exact,
                    limit_best_survival=limit_best_survival,
                    max_structures=max_structures,
                    directory=directory / composition.reduced_formula,
                    # Because we submitted all steady states above, we don't
                    # need the other workflows to do these anymore.
                    singleshot_sources=[],
                    # If set to True, then the current fixed composition will be
                    # written to fixed-compositon-logs
                    ###### OPTIMIZE --- set this to false in the future...?
                    # write_summary_files=False,
                )



# -*- coding: utf-8 -*-

import math
from pathlib import Path

import plotly.graph_objects as plotly_go
from plotly.subplots import make_subplots

from simmate.calculators.vasp.workflows.relaxation.quality00 import (
    Relaxation__Vasp__Quality00,
)
from simmate.calculators.vasp.workflows.relaxation.quality01 import (
    Relaxation__Vasp__Quality01,
)
from simmate.calculators.vasp.workflows.relaxation.quality02 import (
    Relaxation__Vasp__Quality02,
)
from simmate.calculators.vasp.workflows.relaxation.quality03 import (
    Relaxation__Vasp__Quality03,
)
from simmate.calculators.vasp.workflows.relaxation.quality04 import (
    Relaxation__Vasp__Quality04,
)
from simmate.calculators.vasp.workflows.static_energy.quality04 import (
    StaticEnergy__Vasp__Quality04,
)
from simmate.workflow_engine import Workflow

# from simmate.calculators.vasp.database.relaxation import StagedRelaxation


class Relaxation__Vasp__Staged(Workflow):
    """
    Runs a series of increasing-quality relaxations and then finishes with a single
    static energy calculation.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation.vasp.quality00
        - relaxation.vasp.quality01
        - relaxation.vasp.quality02
        - relaxation.vasp.quality03
        - relaxation.vasp.quality04
        - static-energy.vasp.quality04

    This workflow is most useful for randomly-created structures or extremely
    large supercells. More precise relaxations+energy calcs should be done
    afterwards because ettings are still below MIT and Materials Project quality.
    """

    description_doc_short = "runs a series of relaxations (00-04 quality)"
    # database_table = StagedRelaxation

    subworkflows = [
        Relaxation__Vasp__Quality00,
        Relaxation__Vasp__Quality01,
        Relaxation__Vasp__Quality02,
        Relaxation__Vasp__Quality03,
        Relaxation__Vasp__Quality04,
        StaticEnergy__Vasp__Quality04,
    ]

    @classmethod
    def run_config(
        cls,
        structure,
        command=None,
        source=None,
        directory=None,
        copy_previous_directory=False,
        **kwargs,
    ):

        # Our first relaxation is directly from our inputs.
        current_task = cls.subworkflows[0]
        state = current_task.run(
            structure=structure,
            command=command,
            directory=directory / current_task.name_full,
        )
        result = state.result()

        # The remaining tasks continue and use the past results as an input
        for i, current_task in enumerate(cls.subworkflows[1:]):
            preceding_task = cls.subworkflows[i]  # will be one before because of [:1]
            state = current_task.run(
                structure={
                    "database_table": preceding_task.database_table.table_name,
                    "directory": result["directory"],  # uses preceding result
                    "structure_field": "structure_final",
                },
                command=command,
                directory=directory / current_task.name_full,
            )
            result = state.result()

        # we return the final step but update the directory to the parent one
        result["directory"] = directory
        return result

    @classmethod
    def _get_final_energy_series(cls, df, directory: str):
        """
        A utility that grabs the final energies from each stage. Useful for
        grabbing convergence information.
        """
        for subflow in cls.subworkflows:
            pass

    @classmethod
    def get_series_convergence_plot(cls, **filter_kwargs):

        directories = (
            cls.database_table.objects.filter(
                workflow_name=cls.name_full,
                **filter_kwargs,
            )
            .values_list("directory", flat=True)
            .all()
        )

        all_energy_series = []
        for directory in directories:
            energy_series = []
            for subflow in cls.subworkflows:
                query = subflow.database_table.objects.filter(
                    workflow_name=subflow.name_full,
                    directory__startswith=directory,
                    energy_per_atom__isnull=False,
                ).only("energy_per_atom")
                if query.exists():
                    result = query.get()
                    energy_series.append(result.energy_per_atom)
                else:
                    energy_series.append(None)
            all_energy_series.append(energy_series)

        # -------

        figure = make_subplots(
            rows=math.ceil((len(cls.subworkflows) - 1) / 3),
            cols=3,
        )

        for i in range(len(cls.subworkflows) - 1):

            xs = []
            ys = []
            for energy_series in all_energy_series:

                xs.append(energy_series[i + 1])
                ys.append(energy_series[i])

            scatter = plotly_go.Scatter(
                x=xs,
                y=ys,
                mode="markers",
            )
            row = math.ceil((i + 1) / 3)
            col = (i % 3) + 1
            figure.add_trace(
                trace=scatter,
                row=row,
                col=col,
            )

            # Update xaxis properties
            figure.update_xaxes(
                title_text=cls.subworkflows[i + 1].name_full,
                row=row,
                col=col,
            )
            figure.update_yaxes(
                title_text=cls.subworkflows[i].name_full,
                row=row,
                col=col,
            )

        return figure

    @classmethod
    def view_series_convergence_plot(cls, **kwargs):
        figure = cls.get_series_convergence_plot(**kwargs)
        figure.show(renderer="browser")

    @classmethod
    def write_series_convergence_plot(cls, directory: Path, **kwargs):
        figure = cls.get_series_convergence_plot(**kwargs)
        figure.write_html(
            directory / "convergence__staged_relaxations.html",
            include_plotlyjs="cdn",
        )



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .fixed_composition import StructurePrediction__Python__FixedComposition
from .new_individual import StructurePrediction__Python__NewIndividual
from .variable_composition import StructurePrediction__Python__VariableComposition
from .binary_composition import StructurePrediction__Python__BinaryComposition
